@@ -12,7 +12,6 @@
 import re
 
 import netaddr
-from oslo_utils import uuidutils
 
 from ovsdbapp.backend.ovs_idl import command as cmd
 from ovsdbapp.backend.ovs_idl import idlutils
@@ -81,11 +80,12 @@ class LsGetCommand(cmd.BaseGetRowCommand):
     table = 'Logical_Switch'
 
 
-class AclAddCommand(cmd.AddCommand):
+class _AclAddHelper(cmd.AddCommand):
     table_name = 'ACL'
 
-    def __init__(self, api, switch, direction, priority, match, action,
-                 log=False, may_exist=False, **external_ids):
+    def __init__(self, api, entity, direction, priority, match, action,
+                 log=False, may_exist=False, severity=None, name=None,
+                 **external_ids):
         if direction not in ('from-lport', 'to-lport'):
             raise TypeError("direction must be either from-lport or to-lport")
         if not 0 <= priority <= const.ACL_PRIORITY_MAX:
@@ -93,14 +93,16 @@ class AclAddCommand(cmd.AddCommand):
                              const.ACL_PRIORITY_MAX))
         if action not in ('allow', 'allow-related', 'drop', 'reject'):
             raise TypeError("action must be allow/allow-related/drop/reject")
-        super(AclAddCommand, self).__init__(api)
-        self.switch = switch
+        super(_AclAddHelper, self).__init__(api)
+        self.entity = entity
         self.direction = direction
         self.priority = priority
         self.match = match
         self.action = action
         self.log = log
         self.may_exist = may_exist
+        self.severity = severity
+        self.name = name
         self.external_ids = external_ids
 
     def acl_match(self, row):
@@ -109,8 +111,8 @@ class AclAddCommand(cmd.AddCommand):
                 self.match == row.match)
 
     def run_idl(self, txn):
-        ls = self.api.lookup('Logical_Switch', self.switch)
-        acls = [acl for acl in ls.acls if self.acl_match(acl)]
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        acls = [acl for acl in entity.acls if self.acl_match(acl)]
         if acls:
             if self.may_exist:
                 self.result = rowview.RowView(acls[0])
@@ -123,45 +125,183 @@ class AclAddCommand(cmd.AddCommand):
         acl.match = self.match
         acl.action = self.action
         acl.log = self.log
-        ls.addvalue('acls', acl)
+        acl.severity = self.severity
+        acl.name = self.name
+        entity.addvalue('acls', acl)
         for col, value in self.external_ids.items():
             acl.setkey('external_ids', col, value)
         self.result = acl.uuid
 
 
-class AclDelCommand(cmd.BaseCommand):
+class AclAddCommand(_AclAddHelper):
+    lookup_table = 'Logical_Switch'
+
+    def __init__(self, api, switch, direction, priority, match, action,
+                 log=False, may_exist=False, severity=None, name=None,
+                 **external_ids):
+        # NOTE: we're overriding the constructor here to not break any
+        # existing callers before we introduced Port Groups.
+        super(AclAddCommand, self).__init__(api, switch, direction, priority,
+                                            match, action, log, may_exist,
+                                            severity, name, **external_ids)
+
+
+class PgAclAddCommand(_AclAddHelper):
+    lookup_table = 'Port_Group'
+
+
+class _AclDelHelper(cmd.BaseCommand):
+    def __init__(self, api, entity, direction=None,
+                 priority=None, match=None):
+        if (priority is None) != (match is None):
+            raise TypeError("Must specify priority and match together")
+        if priority is not None and not direction:
+            raise TypeError("Cannot specify priority/match without direction")
+        super(_AclDelHelper, self).__init__(api)
+        self.entity = entity
+        self.conditions = []
+        if direction:
+            self.conditions.append(('direction', '=', direction))
+            # priority can be 0
+            if match:  # and therefore priority due to the above check
+                self.conditions += [('priority', '=', priority),
+                                    ('match', '=', match)]
+
+    def run_idl(self, txn):
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        for acl in [a for a in entity.acls
+                    if idlutils.row_match(a, self.conditions)]:
+            entity.delvalue('acls', acl)
+            acl.delete()
+
+
+class AclDelCommand(_AclDelHelper):
+    lookup_table = 'Logical_Switch'
+
+    def __init__(self, api, switch, direction=None,
+                 priority=None, match=None):
+        # NOTE: we're overriding the constructor here to not break any
+        # existing callers before we introduced Port Groups.
+        super(AclDelCommand, self).__init__(api, switch, direction, priority,
+                                            match)
+
+
+class PgAclDelCommand(_AclDelHelper):
+    lookup_table = 'Port_Group'
+
+
+class _AclListHelper(cmd.BaseCommand):
+    def __init__(self, api, entity):
+        super(_AclListHelper, self).__init__(api)
+        self.entity = entity
+
+    def run_idl(self, txn):
+        entity = self.api.lookup(self.lookup_table, self.entity)
+        self.result = [rowview.RowView(acl) for acl in entity.acls]
+
+
+class AclListCommand(_AclListHelper):
+    lookup_table = 'Logical_Switch'
+
+
+class PgAclListCommand(_AclListHelper):
+    lookup_table = 'Port_Group'
+
+
+class QoSAddCommand(cmd.AddCommand):
+    table_name = 'QoS'
+
+    def __init__(self, api, switch, direction, priority, match, rate=None,
+                 burst=None, dscp=None, may_exist=False, **columns):
+        if direction not in ('from-lport', 'to-lport'):
+            raise TypeError("direction must be either from-lport or to-lport")
+        if not 0 <= priority <= const.ACL_PRIORITY_MAX:
+            raise ValueError("priority must be between 0 and %s, inclusive" %
+                             const.ACL_PRIORITY_MAX)
+        if rate is not None and not 1 <= rate <= const.QOS_BANDWIDTH_MAX:
+            raise ValueError("rate(%s) must be between 1 and %s, inclusive" %
+                             rate, const.QOS_BANDWIDTH_MAX)
+        if burst is not None and not 1 <= burst <= const.QOS_BANDWIDTH_MAX:
+            raise ValueError("burst(%s) must be between 1 and %s, "
+                             "inclusive" % burst, const.QOS_BANDWIDTH_MAX)
+        if dscp is not None and not 0 <= dscp <= const.QOS_DSCP_MAX:
+            raise ValueError("dscp(%s) must be between 0 and %s, inclusive" %
+                             dscp, const.QOS_DSCP_MAX)
+        if rate is None and dscp is None:
+            raise ValueError("One of the rate or dscp must be configured")
+        super(QoSAddCommand, self).__init__(api)
+        self.switch = switch
+        self.direction = direction
+        self.priority = priority
+        self.match = match
+        self.rate = rate
+        self.burst = burst
+        self.dscp = dscp
+        self.may_exist = may_exist
+        self.columns = columns
+
+    def qos_match(self, row):
+        return (self.direction == row.direction and
+                self.priority == row.priority and
+                self.match == row.match)
+
+    def run_idl(self, txn):
+        ls = self.api.lookup('Logical_Switch', self.switch)
+        qos_rules = [row for row in ls.qos_rules if self.qos_match(row)]
+        if qos_rules:
+            if self.may_exist:
+                self.result = rowview.RowView(qos_rules[0])
+                return
+            raise RuntimeError("QoS (%s, %s, %s) already exists" % (
+                self.direction, self.priority, self.match))
+        row = txn.insert(self.api.tables[self.table_name])
+        row.direction = self.direction
+        row.priority = self.priority
+        row.match = self.match
+        if self.rate:
+            row.setkey('bandwidth', 'rate', self.rate)
+            if self.burst:
+                row.setkey('bandwidth', 'burst', self.burst)
+        if self.dscp is not None:
+            row.setkey('action', 'dscp', self.dscp)
+        self.set_columns(row, **self.columns)
+        ls.addvalue('qos_rules', row)
+        self.result = row.uuid
+
+
+class QoSDelCommand(cmd.BaseCommand):
     def __init__(self, api, switch, direction=None,
                  priority=None, match=None):
         if (priority is None) != (match is None):
             raise TypeError("Must specify priority and match together")
         if priority is not None and not direction:
             raise TypeError("Cannot specify priority/match without direction")
-        super(AclDelCommand, self).__init__(api)
+        super(QoSDelCommand, self).__init__(api)
         self.switch = switch
         self.conditions = []
         if direction:
             self.conditions.append(('direction', '=', direction))
             # priority can be 0
-            if match:  # and therefor prioroity due to the above check
+            if match:  # and therefor priority due to the above check
                 self.conditions += [('priority', '=', priority),
                                     ('match', '=', match)]
 
     def run_idl(self, txn):
         ls = self.api.lookup('Logical_Switch', self.switch)
-        for acl in [a for a in ls.acls
-                    if idlutils.row_match(a, self.conditions)]:
-            ls.delvalue('acls', acl)
-            acl.delete()
+        for row in ls.qos_rules:
+            if idlutils.row_match(row, self.conditions):
+                ls.delvalue('qos_rules', row)
+                row.delete()
 
 
-class AclListCommand(cmd.BaseCommand):
+class QoSListCommand(cmd.BaseCommand):
     def __init__(self, api, switch):
-        super(AclListCommand, self).__init__(api)
+        super(QoSListCommand, self).__init__(api)
         self.switch = switch
 
     def run_idl(self, txn):
         ls = self.api.lookup('Logical_Switch', self.switch)
-        self.result = [rowview.RowView(acl) for acl in ls.acls]
+        self.result = [rowview.RowView(row) for row in ls.qos_rules]
 
 
 class LspAddCommand(cmd.AddCommand):
@@ -347,7 +487,7 @@ class LspGetPortSecurityCommand(cmd.BaseCommand):
         self.result = lsp.port_security
 
 
-class LspGetUpCommand(cmd.BaseCommand):
+class LspGetUpCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetUpCommand, self).__init__(api)
         self.port = port
@@ -369,7 +509,7 @@ class LspSetEnabledCommand(cmd.BaseCommand):
         lsp.enabled = self.is_enabled
 
 
-class LspGetEnabledCommand(cmd.BaseCommand):
+class LspGetEnabledCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetEnabledCommand, self).__init__(api)
         self.port = port
@@ -391,7 +531,7 @@ class LspSetTypeCommand(cmd.BaseCommand):
         lsp.type = self.port_type
 
 
-class LspGetTypeCommand(cmd.BaseCommand):
+class LspGetTypeCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetTypeCommand, self).__init__(api)
         self.port = port
@@ -412,7 +552,7 @@ class LspSetOptionsCommand(cmd.BaseCommand):
         lsp.options = self.options
 
 
-class LspGetOptionsCommand(cmd.BaseCommand):
+class LspGetOptionsCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetOptionsCommand, self).__init__(api)
         self.port = port
@@ -433,7 +573,7 @@ class LspSetDhcpV4OptionsCommand(cmd.BaseCommand):
         lsp.dhcpv4_options = self.dhcpopt_uuid
 
 
-class LspGetDhcpV4OptionsCommand(cmd.BaseCommand):
+class LspGetDhcpV4OptionsCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LspGetDhcpV4OptionsCommand, self).__init__(api)
         self.port = port
@@ -470,7 +610,7 @@ class DhcpOptionsDelCommand(cmd.BaseCommand):
         dhcpopt.delete()
 
 
-class DhcpOptionsListCommand(cmd.BaseCommand):
+class DhcpOptionsListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r) for
                        r in self.api.tables['DHCP_Options'].rows.values()]
@@ -491,7 +631,7 @@ class DhcpOptionsSetOptionsCommand(cmd.BaseCommand):
         dhcpopt.options = self.options
 
 
-class DhcpOptionsGetOptionsCommand(cmd.BaseCommand):
+class DhcpOptionsGetOptionsCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, dhcpopt_uuid):
         super(DhcpOptionsGetOptionsCommand, self).__init__(api)
         self.dhcpopt_uuid = dhcpopt_uuid
@@ -546,7 +686,7 @@ class LrDelCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LrListCommand(cmd.BaseCommand):
+class LrListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r) for
                        r in self.api.tables['Logical_Router'].rows.values()]
@@ -616,7 +756,7 @@ class LrpDelCommand(PortDelCommand):
             if_exists)
 
 
-class LrpListCommand(cmd.BaseCommand):
+class LrpListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrpListCommand, self).__init__(api)
         self.router = router
@@ -637,7 +777,7 @@ class LrpSetEnabledCommand(cmd.BaseCommand):
         lrp.enabled = self.is_enabled
 
 
-class LrpGetEnabledCommand(cmd.BaseCommand):
+class LrpGetEnabledCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, port):
         super(LrpGetEnabledCommand, self).__init__(api)
         self.port = port
@@ -718,7 +858,7 @@ class LrRouteDelCommand(cmd.BaseCommand):
             raise RuntimeError(msg)
 
 
-class LrRouteListCommand(cmd.BaseCommand):
+class LrRouteListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrRouteListCommand, self).__init__(api)
         self.router = router
@@ -827,7 +967,7 @@ class LrNatDelCommand(cmd.BaseCommand):
                                        match=self.match_ip)
 
 
-class LrNatListCommand(cmd.BaseCommand):
+class LrNatListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrNatListCommand, self).__init__(api)
         self.router = router
@@ -894,7 +1034,7 @@ class LbDelCommand(cmd.BaseCommand):
                 raise
 
 
-class LbListCommand(cmd.BaseCommand):
+class LbListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         self.result = [rowview.RowView(r)
                        for r in self.api.tables['Load_Balancer'].rows.values()]
@@ -939,7 +1079,7 @@ class LrLbDelCommand(cmd.BaseCommand):
             raise
 
 
-class LrLbListCommand(cmd.BaseCommand):
+class LrLbListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, router):
         super(LrLbListCommand, self).__init__(api)
         self.router = router
@@ -988,7 +1128,7 @@ class LsLbDelCommand(cmd.BaseCommand):
             raise
 
 
-class LsLbListCommand(cmd.BaseCommand):
+class LsLbListCommand(cmd.ReadOnlyCommand):
     def __init__(self, api, switch):
         super(LsLbListCommand, self).__init__(api)
         self.switch = switch
@@ -1023,7 +1163,7 @@ class DnsGetCommand(cmd.BaseGetRowCommand):
     table = 'DNS'
 
 
-class DnsListCommand(cmd.BaseCommand):
+class DnsListCommand(cmd.ReadOnlyCommand):
     def run_idl(self, txn):
         table = self.api.tables['DNS']
         self.result = [rowview.RowView(r) for r in table.rows.values()]
@@ -1073,11 +1213,12 @@ class PgAddCommand(cmd.AddCommand):
             try:
                 pg = self.api.lookup(self.table_name, self.name)
                 self.result = rowview.RowView(pg)
+                return
             except idlutils.RowNotFound:
                 pass
 
         pg = txn.insert(self.api._tables[self.table_name])
-        pg.name = self.name
+        pg.name = self.name or ""
         self.set_columns(pg, **self.columns)
         self.result = pg.uuid
 
@@ -1100,40 +1241,34 @@ class PgDelCommand(cmd.BaseCommand):
             raise RuntimeError('Port group %s does not exist' % self.name)
 
 
-class _PgUpdateHelper(cmd.BaseCommand):
+class _PgUpdatePortsHelper(cmd.BaseCommand):
     method = None
 
-    def __init__(self, api, port_group, lsp=None, acl=None, if_exists=False):
-        super(_PgUpdateHelper, self).__init__(api)
+    def __init__(self, api, port_group, lsp=None, if_exists=False):
+        super(_PgUpdatePortsHelper, self).__init__(api)
         self.port_group = port_group
         self.lsp = [] if lsp is None else self._listify(lsp)
-        self.acl = [] if acl is None else self._listify(acl)
         self.if_exists = if_exists
 
     def _listify(self, res):
         return res if isinstance(res, (list, tuple)) else [res]
 
-    def _fetch_resource(self, type_, uuid_):
-        table = 'Logical_Switch_Port' if type_ == 'ports' else 'ACL'
-        return self.api.lookup(table, uuid_)
-
-    def _run_method(self, pg, column, resource):
-        if not resource:
+    def _run_method(self, pg, port):
+        if not port:
             return
 
-        if isinstance(resource, cmd.BaseCommand):
-            resource = resource.result
-        elif uuidutils.is_uuid_like(resource):
+        if isinstance(port, cmd.BaseCommand):
+            port = port.result
+        elif utils.is_uuid_like(port):
             try:
-                resource = self._fetch_resource(column, resource)
+                port = self.api.lookup('Logical_Switch_Port', port)
             except idlutils.RowNotFound:
                 if self.if_exists:
                     return
                 raise RuntimeError(
-                    'Resource %(res)s of type "%(type)s" does not exist' %
-                    {'res': resource, 'type': column})
+                    'Port %s does not exist' % port)
 
-        getattr(pg, self.method)(column, resource)
+        getattr(pg, self.method)('ports', port)
 
     def run_idl(self, txn):
         try:
@@ -1143,15 +1278,16 @@ class _PgUpdateHelper(cmd.BaseCommand):
                                self.port_group)
 
         for lsp in self.lsp:
-            self._run_method(pg, 'ports', lsp)
-
-        for acl in self.acl:
-            self._run_method(pg, 'acls', acl)
+            self._run_method(pg, lsp)
 
 
-class PgAddDataCommand(_PgUpdateHelper):
+class PgAddPortCommand(_PgUpdatePortsHelper):
     method = 'addvalue'
 
 
-class PgDelDataCommand(_PgUpdateHelper):
+class PgDelPortCommand(_PgUpdatePortsHelper):
     method = 'delvalue'
+
+
+class PgGetCommand(cmd.BaseGetRowCommand):
+    table = 'Port_Group'
